import re

class MqttRouter:
    def __init__(self):
        self.routes = {}

    def register_route(self, topic_regex: str, callback: callable):
        if topic_regex not in self.routes:
            self.routes[topic_regex] = []
        self.routes[topic_regex].append(callback)

    def unregister_route(self, topic_regex: str):
        if topic_regex not in self.routes:
            return
        self.routes.pop(topic_regex)

    def inject_message(self, topic, message):
        for route in self.routes:
            result = re.search(route, topic)
            if result is not None:
                for callback in self.routes[route]:
                    callback(topic, message)



from typing import List
import time
import itertools
import os

from Semi_ATE.STDF import MIR, STDR
from ate_apps_common.stdf_utils import (generate_FTR, generate_PIR,
                                        generate_PRR, generate_PTR, generate_SDR,
                                        generate_TSR, generate_HBR,
                                        generate_SBR, generate_MRR,
                                        generate_MIR, generate_PCR,
                                        generate_FAR)


class StdfPartTestContext:
    def __init__(self, site_num: int):
        self.site_num = site_num


STDF_PATH = 'temp_final_stdf_file_on_unload.stdf'


class StdfTestResultAggregator:
    completed_part_records: List[List[STDR]]

    def __init__(self, node_name: str, lot_id: str, job_name: str, sites: list, file_path: str = ''):
        self.version = 'V4'
        self.endian = '<'
        self.completed_part_records = []
        self.num_good_parts = 0
        self.setup_timestamp = int(time.time())
        self.start_timestamp = 0
        self.node_name = node_name
        self.lot_id = lot_id
        self.job_name = job_name
        self.end_timestamp = 0
        self.path = STDF_PATH if not file_path else file_path
        self.sites = sites

        self._clean_up()

    def _clean_up(self):
        if not os.path.exists(self.path):
            return

        os.remove(self.path)

    def finalize(self):
        self.end_timestamp = int(time.time())

    def write_to_file(self, filepath):
        full_file_records = self._stdf_header_records() + list(itertools.chain.from_iterable(self.completed_part_records)) + self._stdf_footer_records()
        with open(filepath, "wb") as f:
            f.write(self.serialize_records(full_file_records))

    def write_records_to_file(self, record):
        with open(self.path, "ab") as f:
            f.write(self.serialize_records(record))

    def write_header_records(self):
        self.set_first_part_test_time()
        self.write_records_to_file(self._stdf_header_records())

    def write_footer_records(self):
        self.write_records_to_file(self._stdf_footer_records())

    def _stdf_header_records(self) -> list:
        return [generate_FAR(2, 4),
                self.generate_MIR(),
                generate_SDR(0, 0, len(self.sites), self.sites)]

    def set_test_program_data(self, test_information: dict):
        self.user_text = test_information['USERTEXT']
        self.program_dir = test_information['PROGRAM_DIR']
        self.temperature = test_information['TEMP']
        self.tester_program = test_information['TESTERPRG']
        self.part_id = test_information['PART_ID']
        self.package_id = test_information['PACKAGE_ID']
        self.sublot_id = test_information['SUBLOT_ID']

    def generate_MIR(self) -> MIR:
        node_num = 81  # TODO: this is not specified yet
        operator_name = 'sct'  # TODO: standard value
        return generate_MIR(self.setup_timestamp, self.start_timestamp, 0,
                            self.lot_id, f'{self.part_id}', self.node_name, f'MSCT-{node_num}',
                            self.job_name, operator_name, self.temperature,
                            self.user_text, f'{self.package_id}', self.sublot_id)

    def get_MIR_dict(self) -> dict:
        return self.generate_MIR().to_dict()

    def _stdf_footer_records(self) -> list:
        return [generate_MRR(self.end_timestamp)]

    def set_first_part_test_time(self):
        self.start_timestamp = int(time.time())

    def append_test_results(self, test_results):
        for test_result in test_results:
            rec = {
                'PIR': lambda test_result: self._generate_PIR(test_result),
                'PTR': lambda test_result: self._generate_PTR(test_result),
                'PRR': lambda test_result: self._generate_PRR(test_result),
                'FTR': lambda test_result: self._generate_FTR(test_result),
            }[test_result['type']](test_result)

            self.write_records_to_file([rec])

    @staticmethod
    def _generate_PRR(prr_record: dict):
        part_fix = [0] * 8
        rec = generate_PRR(prr_record['HEAD_NUM'], prr_record['SITE_NUM'], False, prr_record['NUM_TEST'],
                           prr_record['HARD_BIN'], prr_record['SOFT_BIN'], prr_record['X_COORD'],
                           prr_record['Y_COORD'], prr_record['TEST_T'], prr_record['PART_ID'],
                           prr_record['PART_TXT'], part_fix)

        # TODO: there should not be a is_pass in generate_prr function
        rec.set_value('PART_FLG', prr_record['PART_FLG'])
        return rec

    @staticmethod
    def _generate_PTR(ptr_record: dict) -> dict:
        return generate_PTR(ptr_record['TEST_NUM'], ptr_record['HEAD_NUM'], ptr_record['SITE_NUM'],
                            False, ptr_record['PARM_FLG'], ptr_record['RESULT'], ptr_record['TEST_TXT'],
                            ptr_record['ALARM_ID'], ptr_record['LO_LIMIT'], ptr_record['HI_LIMIT'],
                            ptr_record['UNITS'], ptr_record['C_RESFMT'], ptr_record['RES_SCAL'],
                            ptr_record['LO_SPEC'], ptr_record['HI_SPEC'], ptr_record['OPT_FLAG'])

    @staticmethod
    def _generate_FTR(ftr_record: dict) -> dict:
        rec = generate_FTR(ftr_record['TEST_NUM'], ftr_record['HEAD_NUM'], ftr_record['SITE_NUM'], False)
        rec.set_value('TEST_FLG', ftr_record['TEST_FLG'])
        return rec

    @staticmethod
    def _generate_PIR(pir_record: dict) -> dict:
        return generate_PIR(pir_record['HEAD_NUM'], pir_record['SITE_NUM'])

    def append_test_summary(self, tests_summary: list):
        for test_summary in tests_summary:
            rec = self._generate_TSR(test_summary)
            self.write_records_to_file([rec])

    def append_soft_and_hard_bin_record(self, bin_informations: dict):
        self.write_bin_info(bin_informations,
                            lambda head_num, site_num, bin_num, count, bin_name, bin_pf:
                                generate_SBR(head_num, site_num, bin_num, count, bin_name, bin_pf),
                            0)

        self.write_bin_info(bin_informations,
                            lambda head_num, site_num, bin_num, count, bin_name, bin_pf:
                                generate_HBR(head_num, site_num, bin_num, count, bin_name, bin_pf),
                            1)

    def write_bin_info(self, bin_informations: dict, func: callable, bin_pos: int):
        record_summary = [None] * 6
        for bin_name, bin_info in bin_informations.items():
            record_summary[0] = 255         # head num
            record_summary[1] = 0           # site num
            record_summary[3] = 0           # count
            record_summary[4] = bin_name    # bin name
            record_summary[5] = 'F'         # test flag

            for site_id, bin in bin_info.items():
                self.write_records_to_file([func(bin[4], int(site_id), bin[bin_pos], bin[2], bin_name, bin[3])])
                record_summary[3] += bin[2]
                record_summary[2] = bin[bin_pos]
                record_summary[5] = bin[3]
                record_summary[1] = int(site_id)

            self.write_records_to_file([func(record_summary[0], record_summary[1],
                                             record_summary[2], record_summary[3],
                                             record_summary[4], record_summary[5])])

    def append_part_count_infos(self, part_infos: list):
        pcr_recs = []
        for part_info in part_infos:
            pcr_recs.append(generate_PCR(int(part_info['head_num']), int(part_info['site_num']),
                                         part_info['part_count'], part_info['retest_count'],
                                         part_info['abort_count'], part_info['good_count'],
                                         part_info['functional_count']))

        self.write_records_to_file(pcr_recs)

    @staticmethod
    def _generate_TSR(test_summary: dict) -> dict:
        return generate_TSR(test_summary['HEAD_NUM'], test_summary['SITE_NUM'], test_summary['TEST_TYP'],
                            test_summary['TEST_NUM'], test_summary['EXEC_CNT'], test_summary['FAIL_CNT'],
                            test_summary['ALRM_CNT'], test_summary['TEST_NAM'], test_summary['SEQ_NAME'],
                            test_summary['TEST_LBL'], test_summary['OPT_FLAG'], test_summary['TEST_TIM'],
                            test_summary['TEST_MIN'], test_summary['TEST_MAX'], test_summary['TST_SUMS'],
                            test_summary['TST_SQRS'])

    @staticmethod
    def serialize_records(records) -> bytes:
        return bytes(itertools.chain.from_iterable(rec.__repr__() for rec in records))



from ate_apps_common.transition_sequence import TransitionSequence
from typing import List, Callable


class SequenceContainer:
    def __init__(self, seq: List[str], site_ids: List[str], on_complete: Callable, on_error: Callable):
        self.sequences = {site_id: TransitionSequence(seq[:]) for site_id in site_ids}
        self.on_complete = on_complete
        self.on_error = on_error
        self.done = False

    def trigger_transition(self, site_id: str, transition: str, must_match_transition=True) -> bool:
        seq = self.sequences.get(site_id)
        if seq is None:
            self.on_error(site_id, transition)
            return False
        if not seq.trigger_transition(transition):
            if must_match_transition:
                self.on_error(site_id, transition)
            return False

        if all(x.finished() for x in self.sequences.values()):
            if self.done is False:
                self.on_complete()
                self.done = True

        return True



from Semi_ATE.STDF import (PTR, PRR, PIR, TSR, SBR, HBR, MRR, MIR, PCR, FAR, FTR, SDR)

ENDIAN = '<'


def generate_PTR_dict(test_num, head_num, site_num,
                      is_pass, param_flag, measurement,
                      test_txt, alarm_id, l_limit, u_limit,
                      fmt, exponent, unit, ls_limit, us_limit,
                      opt_flag=2):
    record = {'type': 'PTR'}
    ptr_record = generate_PTR(test_num, head_num, site_num,
                              is_pass, param_flag, measurement,
                              test_txt, alarm_id, l_limit, u_limit,
                              unit, fmt, exponent, ls_limit, us_limit,
                              opt_flag).to_dict()
    ptr_record['TEST_FLG'] = flag_array_to_int(ptr_record['TEST_FLG'], ENDIAN)
    ptr_record['PARM_FLG'] = flag_array_to_int(ptr_record['PARM_FLG'], ENDIAN)
    ptr_record['OPT_FLAG'] = flag_array_to_int(ptr_record['OPT_FLAG'], ENDIAN)
    record.update(ptr_record)
    return record


def generate_PTR(test_num, head_num, site_num,
                 is_pass, param_flag, measurement,
                 test_txt, alarm_id, l_limit, u_limit,
                 unit, fmt, exponent, ls_limit, us_limit,
                 opt_flag=2):
    rec = PTR('V4', ENDIAN)

    format = f'%7{fmt}' if '%7' not in fmt else fmt

    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)
    rec.set_value('TEST_NUM', test_num)

    rec.set_value('TEST_FLG', 0b00000000 if is_pass else 0b00000001)
    rec.set_value('PARM_FLG', param_flag)
    rec.set_value('RESULT', measurement if measurement is not None else ls_limit)
    rec.set_value('TEST_TXT', test_txt)
    rec.set_value('ALARM_ID', alarm_id)

    rec.set_value('C_RESFMT', format)
    rec.set_value('C_LLMFMT', format)
    rec.set_value('C_HLMFMT', format)

    rec.set_value('LO_LIMIT', l_limit)
    rec.set_value('HI_LIMIT', u_limit)

    rec.set_value('LO_SPEC', ls_limit)
    rec.set_value('HI_SPEC', us_limit)

    rec.set_value('RES_SCAL', exponent)
    rec.set_value('LLM_SCAL', exponent)
    rec.set_value('HLM_SCAL', exponent)

    rec.set_value('UNITS', ' ' if '˽' == unit else unit)

    # TODO: workaround, need to explicitly set this or serialization raises exception
    rec.set_value('OPT_FLAG', opt_flag)  # bit set means ignore some field

    # PTR is special in STDF, because all fields AFTER OPT_FLAG of the first PTR in the STDF file for a test defines the default values to be used for all following PTRs of the test
    # possibly not fully implemented here
    return rec


def generate_PIR_dict(head_num, site_num):
    record = {'type': 'PIR'}
    record.update(generate_PIR(head_num, site_num).to_dict())
    return record


def generate_PIR(head_num, site_num):
    rec = PIR('V4', '<')
    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)

    return rec


def generate_PRR(head_num, site_num, is_pass,
                 num_tests, hard_bin, soft_bin,
                 x_coord, y_coord, test_time,
                 part_id, part_txt, part_fix):
    rec = PRR('V4', ENDIAN)

    # 2020-07-31: Force sbin to 1 if it was
    # not set by the user.
    if soft_bin < 0:
        soft_bin = 1

    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)
    rec.set_value('PART_FLG', 0b00000000 if is_pass else 0b00001000)
    rec.set_value('NUM_TEST', num_tests)
    rec.set_value('HARD_BIN', hard_bin)
    rec.set_value('SOFT_BIN', soft_bin)
    rec.set_value('X_COORD', x_coord)
    rec.set_value('Y_COORD', y_coord)
    rec.set_value('TEST_T', test_time)
    rec.set_value('PART_ID', part_id)
    rec.set_value('PART_TXT', part_txt)
    rec.set_value('PART_FIX', part_fix)
    return rec


def generate_PRR_dict(head_num, site_num, is_pass,
                      num_tests, hard_bin, soft_bin,
                      x_coord, y_coord, test_time,
                      part_id, part_txt, part_fix):
    record = {'type': 'PRR'}
    prr_record = generate_PRR(head_num, site_num, is_pass,
                              num_tests, hard_bin, soft_bin,
                              x_coord, y_coord, test_time,
                              part_id, part_txt, part_fix).to_dict()
    prr_record['PART_FLG'] = flag_array_to_int(prr_record['PART_FLG'], ENDIAN)
    prr_record['PART_FIX'] = flag_array_to_int(prr_record['PART_FIX'], ENDIAN)
    record.update(prr_record)
    return record


def generate_TSR(head_num, site_num, test_typ, test_num,
                 exec_cnt, fail_cnt, alarm_cnt, test_nam,
                 seq_name, test_lbl, opt_flag, test_tim,
                 test_min, test_max, tst_sums, tst_sqrs):
    rec = TSR('V4', ENDIAN)

    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)
    rec.set_value('TEST_TYP', test_typ)
    rec.set_value('TEST_NUM', test_num)
    rec.set_value('EXEC_CNT', exec_cnt)
    rec.set_value('FAIL_CNT', fail_cnt)
    rec.set_value('ALRM_CNT', alarm_cnt)
    rec.set_value('TEST_NAM', test_nam)
    rec.set_value('SEQ_NAME', seq_name)
    rec.set_value('TEST_LBL', test_lbl)
    rec.set_value('OPT_FLAG', opt_flag)
    rec.set_value('TEST_TIM', test_tim)
    rec.set_value('TEST_MIN', test_min)
    rec.set_value('TEST_MAX', test_max)
    rec.set_value('TST_SUMS', tst_sums)
    rec.set_value('TST_SQRS', tst_sqrs)

    return rec


def generate_TSR_dict(head_num, site_num, test_typ, test_num,
                      exec_cnt, fail_cnt, alarm_cnt, test_nam,
                      seq_name, test_lbl, opt_flag, test_tim,
                      test_min, test_max, tst_sums, tst_sqrs):
    record = {'type': 'TSR'}
    tsr_record = generate_TSR(head_num, site_num, test_typ, test_num,
                              exec_cnt, fail_cnt, alarm_cnt, test_nam,
                              seq_name, test_lbl, opt_flag, test_tim,
                              test_min, test_max, tst_sums, tst_sqrs).to_dict()
    tsr_record['OPT_FLAG'] = flag_array_to_int(tsr_record['OPT_FLAG'], ENDIAN)
    record.update(tsr_record)
    return record


def flag_array_to_int(flags, endian):
    counter = 1
    if endian == '<':
        counter = -1

    if not flags:
        return

    num = 0
    for index, flag in enumerate(flags[::counter]):
        num += pow(2, index) * int(flag)

    return num


def generate_FTR(test_num, head_num, site_num, exception, opt_flag=255):
    rec = FTR('V4', endian=ENDIAN)
    rec.set_value('TEST_NUM', test_num)
    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)
    rec.set_value('TEST_FLG', 0b10000000 if exception else 0b00000001)
    # RTN_ICNT and PGM_ICNT field must be initialized to satisfy the stdf-record generator
    rec.set_value('RTN_ICNT', 0)
    rec.set_value('PGM_ICNT', 0)
    rec.set_value('OPT_FLAG', opt_flag)
    return rec


def generate_FTR_dict(test_num, head_num, site_num, exception):
    record = {'type': 'FTR'}
    ftr_record = generate_FTR(test_num, head_num, site_num, exception).to_dict()
    ftr_record['TEST_FLG'] = flag_array_to_int(ftr_record['TEST_FLG'], ENDIAN)
    ftr_record['OPT_FLAG'] = flag_array_to_int(ftr_record['OPT_FLAG'], ENDIAN)
    record.update(ftr_record)
    return record


def generate_SBR(head_num: int, site_num: int, bin_num: int, count: int, bin_name: str, bin_pf: str) -> SBR:
    rec = SBR('V4', endian=ENDIAN)
    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)
    rec.set_value('SBIN_NUM', bin_num)
    rec.set_value('SBIN_CNT', count)
    rec.set_value('SBIN_PF', bin_pf)
    rec.set_value('SBIN_NAM', bin_name)
    return rec


def generate_HBR(head_num: int, site_num: int, bin_num: int, count: int, bin_name: str, bin_pf: str) -> HBR:
    rec = HBR('V4', endian=ENDIAN)
    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)
    rec.set_value('HBIN_NUM', bin_num)
    rec.set_value('HBIN_CNT', count)
    rec.set_value('HBIN_PF', bin_pf)
    rec.set_value('HBIN_NAM', bin_name)
    return rec


def generate_MRR(end_timestamp: int) -> MRR:
    rec = MRR('V4', endian=ENDIAN)
    rec.set_value('FINISH_T', end_timestamp)
    return rec


def generate_MIR(setup_time: int, start_time: int, stat_num: int, lot_id: str,
                 part_typ: str, node_name: str, tstr_typ: str, job_name: str,
                 operator_name: str, test_temp: str, user_text: str,
                 package_type: str, sublot_id: str) -> MIR:
    rec = MIR('V4', endian=ENDIAN)

    rec.set_value('SETUP_T', setup_time)
    rec.set_value('START_T', start_time)
    rec.set_value('STAT_NUM', stat_num)
    rec.set_value('MODE_COD', 'P')  # TODO: maybe testprogram can provide this information
    rec.set_value('LOT_ID', lot_id)
    rec.set_value('PART_TYP', part_typ)
    rec.set_value('NODE_NAM', node_name)
    rec.set_value('TSTR_TYP', tstr_typ)
    rec.set_value('JOB_NAM', job_name)
    rec.set_value('SBLOT_ID', sublot_id)
    rec.set_value('OPER_NAM', operator_name)
    rec.set_value('EXEC_VER', job_name)
    rec.set_value('TST_TEMP', test_temp)
    rec.set_value('USER_TXT', user_text)
    rec.set_value('PKG_TYP', package_type)

    return rec


def generate_PCR(head_num: int, site_num: int, part_count: int, retest_count: int,
                 abort_count: int, good_count: int, functional_count: int) -> PCR:
    rec = PCR('V4', endian=ENDIAN)
    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_NUM', site_num)

    rec.set_value('PART_CNT', part_count)
    rec.set_value('RTST_CNT', retest_count)
    rec.set_value('ABRT_CNT', abort_count)
    rec.set_value('GOOD_CNT', good_count)
    rec.set_value('FUNC_CNT', functional_count)

    return rec


def generate_FAR(cpu_type: int, stdf_ver: int) -> FAR:
    rec = FAR('V4', endian=ENDIAN)
    rec.set_value('CPU_TYPE', cpu_type)
    rec.set_value('STDF_VER', stdf_ver)
    return rec


def generate_SDR(head_num: int, site_grp: int, site_cnt: int, site_nums: list) -> SDR:
    rec = SDR('V4', endian=ENDIAN)
    rec.set_value('HEAD_NUM', head_num)
    rec.set_value('SITE_GRP', site_grp)
    rec.set_value('SITE_CNT', site_cnt)
    rec.set_value('SITE_NUM', [int(site_num) for site_num in site_nums])
    return rec



class TransitionSequence:
    def __init__(self, transitions: list):
        self.transitions = transitions[:]
        self.last_transition = ""

    def trigger_transition(self, transition: str) -> bool:
        if self.last_transition == transition:
            return True

        if self.finished():
            return False

        front_transition = self.transitions[0]
        if front_transition != transition:
            return False

        self.last_transition = self.transitions.pop(0)
        return True

    def finished(self):
        return len(self.transitions) == 0



""" Doc """
import json
import os


class ConfigReader:

    """Docstring for ConfigReader. """

    def __init__(self, config_file):
        self.host = None
        self.port = None
        self.config_file = config_file

    def get_configuration(self):
        return self.load_json_from_file(self.config_file)

    def get_configuration_ex(self, *, user_config_file: str = None, user_config_dict: dict = None):
        if self.config_file is not None:
            config = self.get_configuration()
        else:
            config = {}

        if user_config_file is not None:
            user_config = self.load_json_from_file(user_config_file)
            if user_config is None:
                return None

            config.update(user_config)

        if user_config_dict is not None:
            config.update(user_config_dict)

        return config

    def load_json_from_file(self, json_file_path):
        if not os.path.exists(json_file_path):
            raise FileNotFoundError(f'configuration file not found: {json_file_path}')

        with open(json_file_path) as f:
            return json.load(f)



import aiomqtt
import asyncio
import json

from ate_apps_common.mqtt_router import MqttRouter
from ate_common.logger import LogLevel, Logger


class MqttConnection:
    def __init__(self, host: str, port: int, mqtt_client_id: str, logger: Logger):
        self.mqtt_client = aiomqtt.Client(client_id=mqtt_client_id)
        self.mqtt_client.reconnect_delay_set(10, 15)
        self.log = logger
        self.host = host
        self.port = port
        self.router = MqttRouter()
        self.response_event = asyncio.Event()

    def init_mqtt_client_callbacks(self, on_connect, on_disconnect):
        self.mqtt_client.on_connect = on_connect
        self.mqtt_client.on_disconnect = on_disconnect
        self.mqtt_client.on_message = self._on_message_handler

    def start_loop(self):
        self.mqtt_client.connect_async(self.host, self.port)
        self.mqtt_client.loop_start()

    async def stop_loop(self):
        await self.mqtt_client.loop_stop()

    def create_message(self, msg):
        return json.dumps(msg)

    def decode_message(self, message):
        return self.decode_payload(message.payload)

    def decode_payload(self, payload_bytes):
        try:
            payload = json.loads(payload_bytes)
            return payload
        except json.JSONDecodeError as error:
            self.log.log_message(LogLevel.Error(), f'{error}')

        return None

    def set_last_will(self, topic, msg):
        self.mqtt_client.will_set(topic, msg, 2, False)

    def subscribe(self, topic):
        self.mqtt_client.subscribe(topic)

    def publish(self, topic, payload=None, qos=2, retain=False):
        self.mqtt_client.publish(topic, payload, qos, retain)

    def register_route(self, route, callback):
        self.router.register_route(route, callback)

    def subscribe_and_register(self, route, callback):
        self.subscribe(route)
        self.router.register_route(route, callback)

    def unsubscribe(self, topic):
        self.mqtt_client.unsubscribe(topic)

    def unregister_route(self, route):
        self.router.unregister_route(route)

    def _on_message_handler(self, client, userdata, msg):
        self.router.inject_message(msg.topic, msg.payload)

    @staticmethod
    def send_log(_):
        pass

    def response_received(self, topic, data):
        self.response_data = data
        self.response_event.set()

    async def publish_with_response(self, topic_base, data):
        response_topic = f"{topic_base}/response"
        request_topic = f"{topic_base}/request"
        self.response_data = None
        self.response_event.clear()
        self.subscribe_and_register(response_topic, lambda topic, resp: self.response_received(topic, resp))
        self.mqtt_client.publish(request_topic, data)
        await asyncio.wait_for(self.response_event.wait(), 5.0)
        self.unsubscribe(response_topic)
        self.unregister_route(response_topic)
        return self.response_data
